from typing import Dict, Optional

from pydantic import Field

from oogeso.dto import DeviceData
from oogeso.dto.types import ModelType


class DevicePowerSourceData(DeviceData):
    reserve_factor: float = 1  # not used capacity contributes fully to spinning reserve
    model: ModelType = ModelType.POWER_SOURCE


class DevicePowerSinkData(DeviceData):
    model: ModelType = ModelType.POWER_SINK


class DeviceStorageElData(DeviceData):
    E_max: float = 0  # MWh storage capacity
    E_min: float = 0
    E_end: Optional[float] = None  # required storage level at end of horzion
    E_cost: Optional[float] = None  # cost for non-ful storage
    eta: float = 1  # efficiency
    target_profile: Optional[str] = None
    E_init: Optional[float] = 0
    model: ModelType = ModelType.STORAGE_EL


class DeviceSourceElData(DeviceData):
    co2em: Optional[float] = None
    op_cost: Optional[float] = None
    reserve_factor: float = 1  # not used capacity contributes fully to spinning reserve
    model: ModelType = ModelType.SOURCE_EL


class DeviceSourceGasData(DeviceData):
    naturalpressure: float = None
    model: ModelType = ModelType.SOURCE_GAS


class DeviceSourceOilData(DeviceData):
    naturalpressure: float = None
    model: ModelType = ModelType.SOURCE_OIL


class DeviceSourceWaterData(DeviceData):
    naturalpressure: float = None
    model: ModelType = ModelType.SOURCE_WATER


class DeviceSinkElData(DeviceData):
    model: ModelType = ModelType.SINK_EL


class DeviceSinkHeatData(DeviceData):
    model: ModelType = ModelType.SINK_HEAT


class DeviceSinkGasData(DeviceData):
    price: Dict[str, float] = Field(default_factory=lambda: {})
    model: ModelType = ModelType.SINK_GAS


class DeviceSinkOilData(DeviceData):
    price: Dict[str, float] = Field(default_factory=lambda: {})
    model: ModelType = ModelType.SINK_OIL


class DeviceSinkWaterData(DeviceData):
    price: Dict[str, float] = Field(default_factory=lambda: {})
    flow_avg: Optional[float] = None  # required average flow
    max_accumulated_deviation: Optional[float] = None  # buffer size (max accumulated deviation from average)
    model: ModelType = ModelType.SINK_WATER


class DeviceCompressorElData(DeviceData):
    eta: float = None  # efficiency
    Q0: float = None  # nominal flow rate used in linearisation
    temp_in: float = None  # inlet temperature
    model: ModelType = ModelType.COMPRESSOR_EL


class DeviceCompressorGasData(DeviceData):
    eta: float = None  # efficiency
    Q0: float = None  # nominal flow rate used in linearisation
    temp_in: float = None  # inlet temperature
    model: ModelType = ModelType.COMPRESSOR_GAS


class DeviceElectrolyserData(DeviceData):
    eta: float = None  # efficiency
    eta_heat: float = None  # heat recovery efficiency
    model: ModelType = ModelType.ELECTROLYSER


class DeviceFuelCellData(DeviceData):
    eta: float = None  # efficiency
    eta_heat: float = None  # heat recovery efficiency
    model: ModelType = ModelType.FUEL_CELL


class DeviceGasHeaterData(DeviceData):
    eta: float = None  # efficiency
    model: ModelType = ModelType.GAS_HEATER


class DeviceGasTurbineData(DeviceData):
    fuel_A: float = None
    fuel_B: float = None
    eta_heat: float = None
    #    is_on_init: bool = False
    #    startup_cost: float = None
    #    startup_delay: float = None  # Minutes from activation to power delivery
    #    shutdown_cost: float = None
    reserve_factor: float = 1  # not used capacity contributes fully to spinning reserve
    model: ModelType = ModelType.GAS_TURBINE


class DevicePumpData(DeviceData):
    eta: float = None
    model: ModelType = ModelType.PUMP


class DeviceHeatPumpData(DevicePumpData):
    eta: float = None
    model: ModelType = ModelType.HEAT_PUMP


class DevicePumpOilData(DevicePumpData):
    eta: float = None  # efficiency
    model: ModelType = ModelType.PUMP_OIL


class DevicePumpWaterData(DevicePumpData):
    eta: float = None
    model: ModelType = ModelType.PUMP_WATER


class DeviceSeparatorData(DeviceData):
    heat_demand_factor: float = None  # heat demand factor
    model: ModelType = ModelType.SEPARATOR


class DeviceSeparator2Data(DeviceData):
    heat_demand_factor: float = None  # heat demand factor
    model: ModelType = ModelType.SEPARATOR2


class DeviceStorageHydrogenData(DeviceData):
    E_max: float = 0  # MWh storage capacity (maximum stored energy)
    E_min: float = 0
    eta: float = 1  # efficiency
    target_profile: Optional[str] = None  # target profile for use of (seasonal) storage
    E_cost: float = 0  # cost for depleting storage
    E_init: float = 0
    model: ModelType = ModelType.STORAGE_HYDROGEN


class DeviceWellProductionData(DeviceData):
    wellhead_pressure: float = None  # 2 # MPa
    model: ModelType = ModelType.WELL_PRODUCTION


class DeviceWellGasLiftData(DeviceData):
    gas_oil_ratio: float = None  # 500
    water_cut: float = None  # 0.6
    f_inj: float = None  # 220 # gas injection rate as fraction of production rate
    injection_pressure: float = None  # 20 # MPa
    separator_pressure: float = None  # 2 # MPa
    model: ModelType = ModelType.WELL_GAS_LIFT



import pydot


def plot_network(
    simulator,
    timestep=None,
    filename=None,
    prog="dot",
    only_carrier=None,
    rank_direction="LR",
    plot_device_name=False,
    number_format="{:.2g}",
    hide_losses=False,
    hide_edge_label=False,
    **kwargs,
):
    """Plot energy network

    simulator : Simulation object
    timestep : int
        which timestep to show values for
    filename : string
        Name of file
    only_carrier : str or list
        Restrict energy carriers to these types (None=plot all)
    rankdir : str
        Plotting direction TB=top to bottom, LR=left to right
    numberformat : str
        specify how numbers should be represented in plot
    hide_losses : bool
        Don't show losses on edges (if any)
    hide_edgelabel : bool
        Don't show any labels on edges
    **kwargs :
        Additional arguments passed on to pydot.Dot(...)
    """

    # Idea: in general, there are "in" and "out" terminals. If there are
    # no serial devices, then these are merged into a single terminal
    # (prettier plot"). Whether the single terminal is shown as an in or out
    # terminal (left or irght), depends on whether it is an input or output
    # of a majority of the connected devices.

    optimiser = simulator.optimiser
    model = optimiser
    res = simulator.result_object

    col = {
        "t": {
            "el": "red",
            "gas": "orange",
            "heat": "darkgreen",
            "wellstream": "brown",
            "oil": "black",
            "water": "blue4",
            "hydrogen": "deepskyblue2",
        },
        "e": {
            "el": "red",
            "gas": "orange",
            "heat": "darkgreen",
            "wellstream": "brown",
            "oil": "black",
            "water": "blue4",
            "hydrogen": "deepskyblue2",
        },
        "d": "white",
        "cluster": "lightgray",
    }
    # dotG = pydot.Dot(graph_type='digraph') #rankdir='LR',newrank='false')
    dotG = pydot.Dot(graph_type="digraph", rankdir=rank_direction, **kwargs)
    if only_carrier is None:
        carriers = model.setCarrier
    elif type(only_carrier) is str:
        carriers = [only_carrier]
    else:
        carriers = only_carrier

    # plot all node and terminals:
    for n_id, node_obj in optimiser.all_nodes.items():
        cluster = pydot.Cluster(graph_name=n_id, label=n_id, style="filled", color=col["cluster"])
        terms_in = pydot.Subgraph(rank="min")
        gr_devices = pydot.Subgraph(rank="same")
        terms_out = pydot.Subgraph(rank="max")
        for carrier in carriers:
            # add only terminals that are connected to something (device or edge)
            if node_obj.is_non_trivial(carrier):
                devs = node_obj.devices
                num_in = 0
                num_out = 0
                for d, dev_obj in devs.items():
                    dev_model = dev_obj.dev_data.model
                    devlabel = "{}\n{}".format(d, dev_model)
                    if plot_device_name:
                        dev_name = dev_obj.dev_data.name
                        devlabel = "{} {}".format(devlabel, dev_name)
                    carriers_in = dev_obj.carrier_in
                    carriers_out = dev_obj.carrier_out
                    carriers_in_lim = list(set(carriers_in) & set(carriers))
                    carriers_out_lim = list(set(carriers_out) & set(carriers))
                    if (carriers_in_lim != []) or (carriers_out_lim != []):
                        gr_devices.add_node(pydot.Node(d, color=col["d"], style="filled", label=devlabel))
                    if carrier in carriers_in_lim:
                        num_in += 1
                        if timestep is None:
                            devedgelabel = ""
                        else:
                            f_in = res.device_flow[(d, carrier, "in", timestep)]
                            devedgelabel = number_format.format(f_in)
                        if carrier in node_obj.devices_serial:
                            n_in = n_id + "_" + carrier + "_in"
                        else:
                            n_in = n_id + "_" + carrier
                        dotG.add_edge(
                            pydot.Edge(
                                dst=d,
                                src=n_in,
                                color=col["e"][carrier],
                                fontcolor=col["e"][carrier],
                                label=devedgelabel,
                            )
                        )
                    if carrier in carriers_out_lim:
                        num_out += 1
                        if timestep is None:
                            devedgelabel = ""
                        else:
                            f_out = res.device_flow[(d, carrier, "out", timestep)]
                            devedgelabel = number_format.format(f_out)
                        if carrier in node_obj.devices_serial:
                            n_out = n_id + "_" + carrier + "_out"
                        else:
                            n_out = n_id + "_" + carrier
                        dotG.add_edge(
                            pydot.Edge(
                                dst=n_out,
                                src=d,
                                color=col["e"][carrier],
                                fontcolor=col["e"][carrier],
                                label=devedgelabel,
                            )
                        )

                # add in/out terminals
                label_in = carrier  # + "_in "
                label_out = carrier  # + supp + " "
                if timestep is None:
                    pass
                elif carrier in ["gas", "wellstream", "oil", "water"]:
                    label_in += "_" + number_format.format(res.terminal_pressure[(n_id, carrier, "in", timestep)])
                    label_out += "_" + number_format.format(res.terminal_pressure[(n_id, carrier, "out", timestep)])
                elif carrier == "el":
                    if optimiser.all_networks["el"].carrier_data.powerflow_method == "dc-pf":
                        label_in += "_" + number_format.format(res.el_voltage_angle[(n_id, timestep)])
                        label_out += "_" + number_format.format(res.el_voltage_angle[(n_id, timestep)])
                # Add two terminals if there are serial devices, otherwise one:
                if carrier in node_obj.devices_serial:
                    terms_in.add_node(
                        pydot.Node(
                            name=n_id + "_" + carrier + "_in",
                            color=col["t"][carrier],
                            label=label_in,
                            shape="box",
                        )
                    )
                    terms_out.add_node(
                        pydot.Node(
                            name=n_id + "_" + carrier + "_out",
                            color=col["t"][carrier],
                            label=label_out,
                            shape="box",
                        )
                    )
                else:
                    if num_out > num_in:
                        terms_out.add_node(
                            pydot.Node(
                                name=n_id + "_" + carrier,
                                color=col["t"][carrier],
                                label=label_out,
                                shape="box",
                            )
                        )
                    else:
                        terms_in.add_node(
                            pydot.Node(
                                name=n_id + "_" + carrier,
                                color=col["t"][carrier],
                                label=label_out,
                                shape="box",
                            )
                        )

        cluster.add_subgraph(terms_in)
        cluster.add_subgraph(gr_devices)
        cluster.add_subgraph(terms_out)
        dotG.add_subgraph(cluster)

    # plot all edges (per carrier):
    for carrier in carriers:
        for i, edge_obj in optimiser.all_edges.items():
            edge_data = edge_obj.edge_data
            if edge_data.carrier == carrier:
                headlabel = ""
                taillabel = ""
                if hide_edge_label:
                    edgelabel = ""
                elif timestep is None:
                    edgelabel = ""
                    if (not hide_edge_label) and hasattr(edge_data, "pressure_from"):
                        edgelabel = "{} {}-{}".format(
                            edgelabel,
                            edge_data.pressure_from,
                            edge_data.pressure_to,
                        )
                else:
                    edgelabel = number_format.format(res.edge_flow[(i, timestep)])
                    # Add loss
                    if (not hide_losses) and (res.edge_loss is not None) and ((i, timestep) in res.edge_loss):
                        losslabel = number_format.format(res.edge_loss[(i, timestep)])
                        edgelabel = "{} [{}]".format(edgelabel, losslabel)
                n_from = edge_data.node_from
                n_to = edge_data.node_to
                n_from_obj = optimiser.all_nodes[n_from]
                n_to_obj = optimiser.all_nodes[n_to]
                # name of terminal depends on whether it serial or single
                if carrier in n_from_obj.devices_serial:
                    t_out = n_from + "_" + carrier + "_out"
                else:
                    t_out = n_from + "_" + carrier
                if carrier in n_to_obj.devices_serial:
                    t_in = n_to + "_" + carrier + "_in"
                else:
                    t_in = n_to + "_" + carrier

                dotG.add_edge(
                    pydot.Edge(
                        src=t_out,
                        dst=t_in,
                        color='"{0}:invis:{0}"'.format(col["e"][carrier]),
                        fontcolor=col["e"][carrier],
                        label=edgelabel,
                        taillabel=taillabel,
                        headlabel=headlabel,
                    )
                )

    if filename is not None:
        # prog='dot' gives the best layout.
        dotG.write(filename, prog=prog, format="png")
    return dotG



from oogeso.dto.base import (
    CarrierData,
    DeviceData,
    EdgeData,
    EnergySystemData,
    NodeData,
    OptimisationParametersData,
    SimulationResult,
    StartStopData,
    TimeSeriesData,
)
from oogeso.dto.carriers import (
    CarrierElData,
    CarrierFluidData,
    CarrierGasData,
    CarrierHeatData,
    CarrierHydrogenData,
    CarrierOilData,
    CarrierWaterData,
    CarrierWellStreamData,
)
from oogeso.dto.devices import (
    DeviceCompressorElData,
    DeviceCompressorGasData,
    DeviceElectrolyserData,
    DeviceFuelCellData,
    DeviceGasHeaterData,
    DeviceGasTurbineData,
    DeviceHeatPumpData,
    DevicePowerSinkData,
    DevicePowerSourceData,
    DevicePumpData,
    DevicePumpOilData,
    DevicePumpWaterData,
    DeviceSeparator2Data,
    DeviceSeparatorData,
    DeviceSinkElData,
    DeviceSinkGasData,
    DeviceSinkHeatData,
    DeviceSinkOilData,
    DeviceSinkWaterData,
    DeviceSourceElData,
    DeviceSourceGasData,
    DeviceSourceOilData,
    DeviceSourceWaterData,
    DeviceStorageElData,
    DeviceStorageHydrogenData,
    DeviceWellGasLiftData,
    DeviceWellProductionData,
)
from oogeso.dto.edges import (
    EdgeElData,
    EdgeFluidData,
    EdgeGasData,
    EdgeHeatData,
    EdgeHydrogenData,
    EdgeOilData,
    EdgeWaterData,
    EdgeWellstreamData,
)



from .edge import Edge
from .electricity import ElNetwork
from .network import HeatNetwork, HydrogenNetwork, Network
from .network_node import NetworkNode
from .pipelines import GasNetwork, OilNetwork, WaterNetwork, WellStreamNetwork



from enum import Enum


class CarrierType(str, Enum):
    # Fixe: Ensure that this is independent of core!
    EL = "el"
    OIL = "oil"
    WELLSTREAM = "wellstream"
    WATER = "water"
    HEAT = "heat"
    HYDROGEN = "hydrogen"
    FLUID = "fluid"
    GAS = "gas"


class ModelType(str, Enum):
    POWER_SOURCE = "powersource"
    POWER_SINK = "powersink"
    STORAGE_EL = "storageel"
    SOURCE_EL = "sourceel"
    SOURCE_GAS = "sourcegas"
    SOURCE_OIL = "sourceoil"
    SOURCE_WATER = "sourcewater"
    SINK_EL = "sinkel"
    SINK_HEAT = "sinkheat"
    SINK_GAS = "sinkgas"
    SINK_OIL = "sinkoil"
    SINK_WATER = "sinkwater"
    COMPRESSOR_EL = "compressorel"
    COMPRESSOR_GAS = "compressorgas"
    ELECTROLYSER = "electrolyser"
    FUEL_CELL = "fuelcell"
    GAS_HEATER = "gasheater"
    GAS_TURBINE = "gasturbine"
    PUMP = "pump"
    HEAT_PUMP = "heatpump"
    PUMP_OIL = "pumpoil"
    PUMP_WATER = "pumpwater"
    SEPARATOR = "separator"
    SEPARATOR2 = "separator2"
    STORAGE_HYDROGEN = "storagehydrogen"
    WELL_PRODUCTION = "wellproduction"
    WELL_GAS_LIFT = "wellgaslift"


class PressureMethodType(str, Enum):
    WEYMOUTH = "weymouth"
    DARCY_WEISSBACH = "darcy-weissbach"



from typing import Optional

from oogeso import dto
from oogeso.dto.types import PressureMethodType


class CarrierElData(dto.CarrierData):
    powerflow_method: str = "transport"  # "transport","dc_pf"
    reference_node: str = None  # reference node for dc-pf electrical voltage angles
    # required (globally) spinning reserve (MW), -1=no limit
    el_reserve_margin: float = -1
    # required backup margin (MW), -1=no limit
    el_backup_margin: Optional[float] = -1  # MW, -1=no limit
    # minutes, how long stored energy must be sustained to count as reserve:
    reserve_storage_minutes: Optional[int] = None


class CarrierHeatData(dto.CarrierData):
    pass


class CarrierHydrogenData(dto.CarrierData):
    energy_value: float = 13  # MJ/Sm3 (calorific value) -> 13 MJ/Sm3


class CarrierFluidData(dto.CarrierData):
    # Todo: Consider if we need one DTO for each fluid type, or if we can do with one DTO.
    rho_density: float  # kg/m3 -> 900 kg/m3
    viscosity: Optional[float] = None  # kg/(m s) -> 0.0026 kg/(m s)
    G_gravity: Optional[float] = None
    Z_compressibility: Optional[float] = None
    Tb_basetemp_K: Optional[float] = None
    Pb_basepressure_MPa: Optional[float] = None
    pressure_method: Optional[str] = None


class CarrierGasData(CarrierFluidData):
    co2_content: float = None  # kg/Sm3 - see SSB 2016 report -> 2.34 kg/Sm3
    R_individual_gas_constant: float = None  # J/(kg K) -> 500 J/kgK
    energy_value: float = None  # MJ/Sm3 (calorific value) -> 40 MJ/Sm3
    k_heat_capacity_ratio: float = None  # 1.27
    pressure_method: PressureMethodType = PressureMethodType.WEYMOUTH  # pressure drop calculation


class CarrierWellStreamData(CarrierFluidData):
    darcy_friction: float = None  # 0.02
    water_cut: float = None
    gas_oil_ratio: float = None


class CarrierOilData(CarrierFluidData):
    darcy_friction: float = None  # 0.02
    pressure_method: PressureMethodType = PressureMethodType.DARCY_WEISSBACH  # pressure drop calculation


class CarrierWaterData(CarrierFluidData):
    darcy_friction: float = None  # 0.02
    pressure_method: PressureMethodType = PressureMethodType.DARCY_WEISSBACH  # pressure drop calculation



from typing import List, Optional, Tuple

import pandas as pd
from pydantic import BaseModel, Field, Extra

from oogeso.dto.types import CarrierType, ModelType


class StartStopData(BaseModel, extra=Extra.forbid):
    is_on_init: bool = False  # Initial on/off status
    penalty_start: float = 0  # Startup "cost"
    penalty_stop: float = 0  # Shutdown "cost"
    delay_start_minutes: int = 0  # Delay in minutes from activation to online
    minimum_time_on_minutes: float = 0  # Minimum on-time in minutes once started
    minimum_time_off_minutes: float = 0  # Minimum off-time in minutes once stopped


class TimeSeriesData(BaseModel, extra=Extra.forbid):
    id: str
    data: List[float]
    data_nowcast: Optional[List[float]] = None


class OptimisationParametersData(BaseModel, extra=Extra.forbid):
    # name of objective function to use:
    objective: str
    # minutes per timestep:
    time_delta_minutes: int
    # timesteps in each rolling optimisation:
    planning_horizon: int
    # timesteps between each optimisation:
    optimisation_timesteps: int
    # timesteps beyond which forecast (instead of nowcast) profile is used:
    forecast_timesteps: int
    # costs for co2 emissions (currency/kgCO2)
    co2_tax: Optional[float] = None
    # limit on allowable emission intensity (kgCO2/Sm3oe), -1=no limit
    emission_intensity_max: Optional[float] = -1
    # limit on allowable emission intensity (kgCO2/hour), -1= no limit
    emission_rate_max: Optional[float] = -1
    # how to represent piecewise linear constraints:
    piecewise_repn: str = "SOS2"
    # specify which data to return from simulation as a list. None=return all data
    optimisation_return_data: Optional[List[str]] = None


class CarrierData(BaseModel, extra=Extra.forbid):
    id: str


class DeviceData(BaseModel, extra=Extra.forbid):  # Parent class - use subclasses instead
    id: str  # unique identifier
    node_id: str  # reference to node identifier
    name: str = ""
    include: Optional[bool] = True
    profile: Optional[str] = None  # reference to time-series
    flow_min: Optional[float] = None  # Energy or fluid flow limit
    flow_max: Optional[float] = None
    # Ramp rates are given as change relative to capacity per minute, 1=100%/min:
    max_ramp_down: Optional[float] = None
    max_ramp_up: Optional[float] = None
    start_stop: Optional[StartStopData] = None
    reserve_factor: float = 0  # contribution to electrical spinning reserve
    op_cost: Optional[float] = None
    # Penalty may be fuel, emissions, cost and combinations of these
    penalty_function: Optional[Tuple[List[float], List[float]]] = None
    model: ModelType


class EdgeData(BaseModel, extra=Extra.forbid):  # Base model, use implementations below
    id: str
    node_from: str
    node_to: str
    length_km: Optional[float] = None
    flow_max: float = None  # Maximum flow (MW or Sm3/s)
    bidirectional: Optional[bool] = True
    include: bool = True  # whether to include object in problem formulation
    carrier: CarrierType


class NodeData(BaseModel, extra=Extra.forbid):
    # unique identifier:
    id: str


class SimulationResult(BaseModel, extra=Extra.forbid, arbitrary_types_allowed=True):
    """Results from oogeso simulation

    The results are stored in a set of multi-index Series, with
    index names indicating what they are:

    device - device identifier
    node - node identifier
    edge - edge identifier
    carrier - network type ("el", "gas", "oil", "water", "hydrogen", "heat")
    terminal - input/output ("in" or "out"),
    time (integer timestep)
    """

    # Input/output flow per device and network type:
    device_flow: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Device startup preparation status (boolean):
    device_is_prep: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Device on/off status (boolean):
    device_is_on: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Device starting status (boolean):
    device_starting: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Device stopping status (boolean):
    device_stopping: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Energy storage filling level (Sm3 or MJ)
    device_storage_energy: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Max available "flow" (power/fluid) from storage (Sm3/s or MW):
    device_storage_pmax: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Device assosiated penalty rate (PENALTY_UNIT/s):
    penalty: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Flow rate (Sm3/s or MW):
    edge_flow: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Loss rate (MW) - only relevant for energy flow (el and heat):
    edge_loss: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Voltage angle at node - only relevant for electricity floc computed via dc-pf:
    el_voltage_angle: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Pressure at node (MPa):
    terminal_pressure: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Direct flow between in and out terminal of node - relevant if there is no device inbetween:
    terminal_flow: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Emission rate (sum of all devices) (kgCO2/s):
    co2_rate: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Emission rate per device (kgCO2/s):
    co2_rate_per_dev: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Revenue rate for exported oil/gas (CURRENCY/s):
    export_revenue: Optional[pd.Series] = Field(default_factory=pd.Series)
    # CO2 intensity of exported oil/gas (kgCO2/Sm3oe):
    co2_intensity: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Available online electrical reserve capacity (MW):
    el_reserve: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Available online electrical backup per device (MW):
    el_backup: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Value of duals (associated with constraints)
    duals: Optional[pd.Series] = Field(default_factory=pd.Series)
    # Time-series profiles used in simulation (copied from the input)
    profiles_forecast: Optional[pd.DataFrame] = Field(default_factory=pd.DataFrame)
    profiles_nowcast: Optional[pd.DataFrame] = Field(default_factory=pd.DataFrame)

    def append_results(self, sim_res):
        exclude_list = ["df_profiles_forecast", "df_profiles_forecast"]
        for field_name in self.__fields__:
            if field_name not in exclude_list:
                my_df = getattr(self, field_name)
                other_df = getattr(sim_res, field_name)
                if other_df is not None:
                    new_df = pd.concat([my_df, other_df]).sort_index()
                    if isinstance(new_df.index[0], tuple):  # Fix for multi-index DataFrame and Series
                        new_df.index = pd.MultiIndex.from_tuples(new_df.index)
                    new_df.index.names = other_df.index.names
                    setattr(self, field_name, new_df)


class EnergySystemData(BaseModel, extra=Extra.forbid):
    parameters: OptimisationParametersData
    carriers: List[CarrierData]
    nodes: List[NodeData]
    edges: List[EdgeData]
    devices: List[DeviceData]
    profiles: List[TimeSeriesData]



from typing import Callable

from oogeso import dto
from oogeso.core import devices, networks


def get_network_from_carrier_name(carrier_name: str) -> Callable:
    map_carrier_to_model = {
        "el": networks.ElNetwork,
        "gas": networks.GasNetwork,
        "oil": networks.OilNetwork,
        "water": networks.WaterNetwork,
        "wellstream": networks.WellStreamNetwork,
        "heat": networks.HeatNetwork,
        "hydrogen": networks.HydrogenNetwork,
    }
    if not carrier_name.lower() in map_carrier_to_model.keys():
        raise NotImplementedError(f"Carrier: {carrier_name} has not implemented a corresponding Network mapping")
    return map_carrier_to_model[carrier_name.lower()]


def get_device_from_model_name(model_name: str) -> Callable:
    map_device_name_to_class = {
        "powersource": devices.PowerSource,
        "powersink": devices.PowerSink,
        "storageel": devices.StorageEl,
        "compressorel": devices.CompressorEl,
        "compressorgas": devices.CompressorGas,
        "electrolyser": devices.Electrolyser,
        "fuelcell": devices.FuelCell,
        "gasheater": devices.GasHeater,
        "gasturbine": devices.GasTurbine,
        "heatpump": devices.HeatPump,
        "pumpoil": devices.PumpOil,
        "pumpwater": devices.PumpWater,
        "separator": devices.Separator,
        "separator2": devices.Separator2,
        "sinkel": devices.SinkEl,
        "sinkheat": devices.SinkHeat,
        "sinkgas": devices.SinkGas,
        "sinkoil": devices.SinkOil,
        "sinkwater": devices.SinkWater,
        "sourceel": devices.SourceEl,
        "sourcegas": devices.SourceGas,
        "sourceoil": devices.SourceOil,
        "sourcewater": devices.SourceWater,
        "storagehydrogen": devices.StorageHydrogen,
        "wellgaslift": devices.WellGasLift,
        "wellproduction": devices.WellProduction,
    }
    if model_name in map_device_name_to_class:
        return map_device_name_to_class[model_name]
    else:
        raise NotImplementedError(f"Device {model_name} has not been implemented.")


def get_device_data_class_from_str(model_name: str) -> Callable:
    model_name = model_name.replace("_", "")
    map_device_name_to_class = {
        "powersource": dto.DevicePowerSourceData,
        "powersink": dto.DevicePowerSinkData,
        "storageel": dto.DeviceStorageElData,
        "compressorel": dto.DeviceCompressorElData,
        "compressorgas": dto.DeviceCompressorGasData,
        "electrolyser": dto.DeviceElectrolyserData,
        "fuelcell": dto.DeviceFuelCellData,
        "gasheater": dto.DeviceGasHeaterData,
        "gasturbine": dto.DeviceGasTurbineData,
        "heatpump": dto.DeviceHeatPumpData,
        "pumpoil": dto.DevicePumpOilData,
        "pumpwater": dto.DevicePumpWaterData,
        "separator": dto.DeviceSeparatorData,
        "separator2": dto.DeviceSeparator2Data,
        "sinkel": dto.DeviceSinkElData,
        "sinkheat": dto.DeviceSinkHeatData,
        "sinkgas": dto.DeviceSinkGasData,
        "sinkoil": dto.DeviceSinkOilData,
        "sinkwater": dto.DeviceSinkWaterData,
        "sourceel": dto.DeviceSourceElData,
        "sourcegas": dto.DeviceSourceGasData,
        "sourceoil": dto.DeviceSourceOilData,
        "sourcewater": dto.DeviceSourceWaterData,
        "storagehydrogen": dto.DeviceStorageHydrogenData,
        "wellgaslift": dto.DeviceWellGasLiftData,
        "wellproduction": dto.DeviceWellProductionData,
    }
    if model_name in map_device_name_to_class:
        return map_device_name_to_class[model_name]
    else:
        raise NotImplementedError(f"Device data class for {model_name} has not been implemented.")


def get_carrier_data_class_from_str(model_name: str) -> Callable:
    map_carrier_name_to_class = {
        "el": dto.CarrierElData,
        "gas": dto.CarrierGasData,
        "oil": dto.CarrierOilData,
        "water": dto.CarrierWaterData,
        "hydrogen": dto.CarrierHydrogenData,
        "heat": dto.CarrierHeatData,
        "wellstream": dto.CarrierWellStreamData,
    }
    if model_name in map_carrier_name_to_class:
        return map_carrier_name_to_class[model_name]
    else:
        raise NotImplementedError(f"Carrier data class for {model_name} has not been implemented.")


def get_edge_data_class_from_str(carrier_name: str) -> Callable:
    map_edge_name_to_class = {
        "el": dto.EdgeElData,
        "gas": dto.EdgeGasData,
        "oil": dto.EdgeOilData,
        "water": dto.EdgeWaterData,
        "hydrogen": dto.EdgeHydrogenData,
        "heat": dto.EdgeHeatData,
        "wellstream": dto.EdgeWellstreamData,
    }
    if carrier_name in map_edge_name_to_class:
        return map_edge_name_to_class[carrier_name]
    else:
        raise NotImplementedError(f"Edge data class for {carrier_name} has not been implemented.")



import logging
from typing import List, Optional

import pandas as pd

from oogeso import dto

logger = logging.getLogger(__name__)


def create_time_series_data(
    df_forecast: pd.DataFrame,
    df_nowcast: pd.DataFrame,
    time_start: Optional[str],
    time_end: Optional[str],
    timestep_minutes: int,
    resample_method: str = "linear",
) -> List[dto.TimeSeriesData]:
    """Rearrange and resample pandas timeseries to Oogeso data transfer object

    The input dataframes should have a datetime index
    """

    # Compine forecast and nowcast timeseries into a single dataframe
    df_orig = pd.concat({"forecast": df_forecast, "nowcast": df_nowcast}, axis=1)
    # 1 resample
    resampled = df_orig.resample(
        rule="{}T".format(
            timestep_minutes,
        ),
        closed="left",
        label="left",
    )
    df_new = resampled.mean()

    # If up-sampling, NaN values filed by linear interpolation (or other method)
    # If down-sampling, this does nothing
    df_new = df_new.interpolate(method=resample_method)

    # Select the times of interest
    if time_start is not None:
        mask = df_new.index >= time_start
        df_new = df_new[mask]
    if time_end is not None:
        mask = df_new.index <= time_end
        df_new = df_new[mask]
    if df_new.isna().any().any():
        logger.warning("Profiles contain NA")

    # now, create Oogeso TimeSereriesData object
    profiles = []
    for col in df_new.columns:
        if col[0] == "forecast":
            curve = col[1]
            list_data = df_new.loc[:, ("forecast", curve)].to_list()
            list_data_nowcast = None
            if ("nowcast", curve) in df_new.columns:
                list_data_nowcast = df_new.loc[:, ("nowcast", curve)].to_list()
            new_ts = dto.TimeSeriesData(id=curve, data=list_data, data_nowcast=list_data_nowcast)
            profiles.append(new_ts)
        elif col[0] == "nowcast":
            if ("forecast", col[1]) not in df_new.columns:
                logger.warning(f"Nowcast but no forecast profile for {col[1]}")

    return profiles



import logging
from typing import Dict, Optional, Tuple, Union

import numpy as np
import pyomo.environ as pyo
import scipy

from oogeso import dto
from oogeso.core.networks.edge import FluidEdge
from oogeso.core.networks.network import Network

logger = logging.getLogger(__name__)

GRAVITY_ACCELERATION_CONSTANT = 9.8  # m/s^2


class FluidNetwork(Network):
    carrier_data: dto.CarrierFluidData
    edges: Dict[str, FluidEdge]

    def define_constraints(self, pyomo_model: pyo.Model):
        super().define_constraints(pyomo_model)

        # additional
        if self.carrier_data.pressure_method in ["weymouth", "darcy-weissbach"]:
            edgelist = [e for e in self.edges.keys()]
            if edgelist:
                # non-empty list of edges
                constr_flow = pyo.Constraint(
                    edgelist,
                    pyomo_model.setHorizon,
                    rule=self._rule_pipeline_flow,
                )
                setattr(
                    pyomo_model,
                    "constrE_{}_{}".format(self.carrier_id, "flow"),
                    constr_flow,
                )

    def _rule_pipeline_flow(self, model: pyo.Model, edge, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """Pipeline flow vs pressure drop"""
        # edge = self.id
        edge_obj = self.edges[edge]
        edge_data = edge_obj.edge_data
        carrier = edge_data.carrier
        n_from = edge_data.node_from
        n_to = edge_data.node_to
        print_log = True if t == 0 else False

        p1 = model.varPressure[(n_from, carrier, "out", t)]
        p2 = model.varPressure[(n_to, carrier, "in", t)]
        Q = model.varEdgeFlow[edge, t]
        if hasattr(edge_data, "num_pipes"):
            if edge_data.num_pipes is not None:
                num_pipes = edge_data.num_pipes
                if print_log:
                    logger.debug("{},{}: {} parallel pipes".format(edge, t, num_pipes))
                Q = Q / num_pipes
        else:
            raise ValueError("FluidNetworkEdge is expected to have the attribute num_pipes.")
        p2_computed = self.compute_edge_pressuredrop(edge_obj, p1=p1, Q=Q, linear=True, print_log=print_log)
        return p2 == p2_computed

    @staticmethod
    def _compute_exps_and_k(edge_data: dto.EdgeFluidData, carrier_data: dto.CarrierFluidData):
        """Derive exp_s and k parameters for Weymouth equation"""
        # gas pipeline parameters - derive k and exp(s) parameters:
        ga = carrier_data

        # if 'temperature_K' in model.paramEdge[edge]:
        temp = edge_data.temperature_K
        height_difference = edge_data.height_m
        length = edge_data.length_km
        diameter = edge_data.diameter_mm
        s = 0.0684 * (ga.G_gravity * height_difference / (temp * ga.Z_compressibility))
        if s > 0:
            # height difference - use equivalent length
            sfactor = (np.exp(s) - 1) / s
            length = length * sfactor

        k = (
            4.3328e-8
            * ga.Tb_basetemp_K
            / ga.Pb_basepressure_MPa
            * (ga.G_gravity * temp * length * ga.Z_compressibility) ** (-1 / 2)
            * diameter ** (8 / 3)
        )
        exp_s = np.exp(s)
        return exp_s, k

    def compute_edge_pressuredrop(
        self, edge: FluidEdge, p1, Q, method: Optional[str] = None, linear: bool = False, print_log: bool = True
    ):
        """Compute pressure drop in pipe

        parameters
        ----------
        edge : edge object
        p1 : float
            pipe inlet pressure (MPa)
        Q : float
            flow rate (Sm3/s)
        method : string
            None, weymouth, darcy-weissbach
        linear : boolean
            whether to use linear model or not
        print_log: bool
            To print or not to print the log

        Returns
        -------
        p2 : float
            pipe outlet pressure (MPa)"""

        # edge = self.id
        edge_data = edge.edge_data
        edge_id = edge_data.id
        carrier = edge_data.carrier
        carrier_data = self.carrier_data
        if carrier_data.pressure_method is not None:
            method = carrier_data.pressure_method

        n_from = edge_data.node_from
        n_to = edge_data.node_to
        p0_from = edge.node_from.get_pressure_nominal(carrier, "out")
        p0_to = edge.node_to.get_pressure_nominal(carrier, "in")
        if (p0_from is not None) and (p0_to is not None):
            if linear & (p0_from == p0_to):
                method = None
                if print_log:
                    logger.debug(f"{n_from}-{n_to}: Pipe without pressure drop" " ({p0_from} / {p0_to} MPa)")
        elif linear:
            # linear equations, but nominal values not given - assume no drop
            # logger.debug("{}-{}: Aassuming no  pressure drop".format(n_from, n_to))
            method = None
        else:
            # use non-linear equations, no nominal pressure required
            pass

        if method is None:
            # no pressure drop
            p2 = p1
            return p2

        elif method == "weymouth":
            """
            Q = k * sqrt( Pin^2 - e^s Pout^2 ) [Weymouth equation, nonlinear]
                => Pout = sqrt(Pin^2-Q^2/k^2)/e^2
            Q = c * (Pin_0 Pin - e^s Pout0 Pout) [linearised version]
                => Pout = (Pin0 Pin - Q/c)/(e^s Pout0)
            c = k/sqrt(Pin0^2 - e^s Pout0^2)

            REFERENCES:
            1) E Sashi Menon, Gas Pipeline Hydraulics, Taylor & Francis (2005),
            https://doi.org/10.1201/9781420038224
            2) A Tomasgard et al., Optimization  models  for  the  natural  gas
            value  chain, in: Geometric Modelling, Numerical Simulation and
            Optimization. Springer Verlag, New York (2007),
            https://doi.org/10.1007/978-3-540-68783-2_16
            """
            exp_s, k = self._compute_exps_and_k(edge_data, carrier_data=carrier_data)
            if print_log:
                logger.debug("pipe {}: exp_s={}, k={}".format(edge_id, exp_s, k))
            if linear:
                p_from = p1
                x0 = p0_from ** 2 - exp_s * p0_to ** 2
                coeff = k * x0 ** (-1 / 2)
                p2 = (p0_from * p_from - Q / coeff) / (exp_s * p0_to)
            else:
                # weymouth eqn (non-linear)
                p2 = 1 / exp_s * (p1 ** 2 - Q ** 2 / k ** 2) ** (1 / 2)
            return p2

        elif method == "darcy-weissbach":
            rho = carrier_data.rho_density
            D = edge_data.diameter_mm / 1000
            L = edge_data.length_km * 1000
            height_difference = edge_data.height_m

            if (carrier_data.viscosity is not None) & (not linear):
                # compute darcy friction factor from flow rate and viscosity (nonlinear)
                mu = carrier_data.viscosity
                Re = 2 * rho * Q / (np.pi * mu * D)
                f = 1 / (0.838 * scipy.special.lambertw(0.629 * Re)) ** 2
                f = f.real
            elif hasattr(carrier_data, "darcy_friction") and getattr(carrier_data, "darcy_friction") is not None:
                f = carrier_data.darcy_friction
            else:
                raise Exception("Must provide viscosity or darcy_friction for {}".format(carrier))
            (p2, Q0) = darcy_weissbach_p2(
                Q,
                p1 * 1e6,
                f=f,
                rho=rho,
                diameter=D,
                length=L,
                height_difference=height_difference,
                linear=linear,
                p0_from=p0_from * 1e6,
                p0_to=p0_to * 1e6,
            )
            p2 = p2 * 1e-6  # Convert to MPa
            if print_log:
                logger.debug(f"derived pipe ({edge_id}) flow rate:' ' Q={Q}, linearQ0={Q0:5.3g}, ' ' friction={f:5.3g}")

            return p2
        else:
            raise Exception("Unknown pressure drop calculation method ({})".format(method))


class GasNetwork(FluidNetwork):
    pass


class OilNetwork(FluidNetwork):
    pass


class WaterNetwork(FluidNetwork):
    pass


class WellStreamNetwork(FluidNetwork):
    pass


def darcy_weissbach_p2(
    Q,
    p1,
    f,
    rho,
    diameter,
    length,
    height_difference=0,
    linear=True,
    p0_from=None,  # Pa
    p0_to=None,  # Pa
) -> Tuple[float, ...]:
    """compute outlet pressure from darcy-weissbach equation

    parameters
    ----------
    p1 : float
        pressure at pipe inlet (Pa)
    Q : float
        flow rate (Sm3/s)
    f : float
        friction factor
    rho : float
        fluid density (kg/m3)
    diameter : float
        pipe inner diameter (m)
    length : float
        pipe length (m)
    height_difference : float
        height difference outlet vs inlet (m)
    linear : True or False
    p0_from : Optional pressure from
    p0_to : Optional pressure to

    """
    D = diameter  # m
    L = length  # m
    grav = GRAVITY_ACCELERATION_CONSTANT
    q0: float = np.nan
    if linear:
        k2 = (8 * f * rho * L) / (np.pi ** 2 * D ** 5)
        q0_squared = 1 / k2 * ((p0_from - p0_to) - rho * grav * height_difference)
        if q0_squared < 0:
            logging.error(
                "Negative q0^2 (flow direction error) q0^2=%s p1=%s p2=%s rho=%s grav=%s z=%s",
                q0_squared,
                p0_from,
                p0_to,
                rho,
                grav,
                height_difference,
            )
        q0 = np.sqrt(q0_squared)
        p2: float = p1 - rho * grav * height_difference - k2 * (2 * q0 * Q - q0 ** 2)
    else:
        # Quadratic in Q
        p2: float = p1 - rho * grav * height_difference - 8 * f * rho * L * Q ** 2 / (np.pi ** 2 * D ** 5)
    return tuple([p2, q0])


def darcy_weissbach_Q(p1, p2, f, rho, diameter_mm, length_km, height_difference_m=0):
    """compute flow rate from non-linear darcy-weissbach eqn

    parameters
    ----------
    p1 : float
        pressure at pipe input (MPa)
    p2 : float
        pressure at pipe output (MPa)
    f : float
        friction factor
    rho : float
        fluid density (kg/m3)
    diameter_mm : float
        pipe inner diameter (mm)
    length_km : float
        pipe length (km)
    height_difference_m : float
        height difference output vs input (m)

    """

    grav = 9.98
    L = length_km * 1000
    D = diameter_mm / 1000
    k = 8 * f * rho * L / (np.pi ** 2 * D ** 5)
    Q = np.sqrt(((p1 - p2) * 1e6 - rho * grav * height_difference_m) / k)
    return Q



import logging
from typing import Dict, Optional, Tuple, Union

import pyomo.environ as pyo

from oogeso import dto
from oogeso.core.devices.base import Device
from oogeso.core.networks import electricalsystem as el_calc
from oogeso.core.networks.edge import ElEdge
from oogeso.core.networks.network import Network

logger = logging.getLogger(__name__)


class ElNetwork(Network):
    def __init__(
        self,
        carrier_data: dto.CarrierElData,
        edges: Dict[str, ElEdge],
    ):
        super().__init__(carrier_data=carrier_data, edges=edges)
        self.carrier_data = carrier_data
        self.edges = edges
        self.el_flow_coeff_B: Optional[Dict[Tuple[str, str], float]] = None
        self.el_flow_coeff_DA: Optional[Dict[Tuple[str, str], float]] = None

    def define_constraints(self, pyomo_model: pyo.Model) -> None:
        super().define_constraints(pyomo_model=pyomo_model)

        if self.carrier_data.powerflow_method == "dc-pf":
            raise NotImplementedError("Power flow equation not implemented yet.")

            logger.warning("Code for electric powerflow calculations need improvement (pu conversion)")
            nodelist = list(pyomo_model.setNode)  # self.all_nodes.keys()
            edgelist_el = {edge_id: edge.edge_data.dict() for edge_id, edge in self.edges.items()}
            coeff_B, coeff_DA = el_calc.compute_power_flow_matrices(nodelist, edgelist_el, base_Z=1)
            self.el_flow_coeff_B = coeff_B
            self.el_flow_coeff_DA = coeff_DA

            # Reference voltage node:
            constr_el_voltage_reference = pyo.Constraint(pyomo_model.setHorizon, rule=self._rule_el_voltage_reference)
            setattr(
                pyomo_model,
                "constrN_{}_{}".format(self.carrier_data.reference_node, "voltageref"),
                constr_el_voltage_reference,
            )
            # Linearised power flow equations:
            for edge in self.edges.keys():
                constr_flow = pyo.Constraint(
                    edge,
                    pyomo_model.setHorizon,
                    rule=self._rule_dc_power_flow_equations,
                )
                setattr(pyomo_model, "constrE_{}_{}".format(edge, "flow"), constr_flow)

    def _rule_el_voltage_reference(self, pyomo_model: pyo.Model, t: int) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        el_carrier = self.carrier_data
        n = el_carrier.reference_node
        expr = pyomo_model.varElVoltageAngle[n, t] == 0
        return expr

    def _rule_dc_power_flow_equations(
        self, pyomo_model: pyo.Model, edge: str, t: int
    ) -> Union[pyo.Expression, pyo.Constraint.Skip]:
        """Flow vs voltage angle"""
        base_mva = el_calc.elbase["baseMVA"]
        base_angle = el_calc.elbase["baseAngle"]
        lhs = pyomo_model.varEdgeFlow[edge, t]
        lhs = lhs / base_mva

        n2s = [k[1] for k in self.el_flow_coeff_DA.keys() if k[0] == edge]

        rhs = sum([self.el_flow_coeff_DA[edge, n2] * (pyomo_model.varElVoltageAngle[n2, t] * base_angle)] for n2 in n2s)
        return lhs == rhs

    @staticmethod
    def compute_el_reserve(
        pyomo_model,
        t,
        all_devices: Dict[str, Device],
        exclude_device: str = None,
    ):
        """compute non-used generator capacity (and available loadflex)
        This is reserve to cope with forecast errors, e.g. because of wind
        variation or motor start-up

        all_devices : dict
            dictionary of all devices in the model {device_id:device_object}
        exclude_device : str (default None)
            compute reserve by devices excluding this one
        """
        alldevs = [d for d in pyomo_model.setDevice if d != exclude_device]
        # relevant devices are devices with el output or input
        cap_avail = 0.0
        p_generating = 0.0
        loadreduction = 0.0
        for d in alldevs:
            dev_obj = all_devices[d]
            reserve = dev_obj.compute_el_reserve(pyomo_model, t)
            cap_avail += reserve["capacity_available"]
            p_generating += reserve["capacity_used"]
            loadreduction += reserve["loadreduction_available"]

        res_dev = (cap_avail - p_generating) + loadreduction
        return res_dev



from typing import List, Optional, Tuple, Union

from oogeso.dto import EdgeData
from oogeso.dto.types import CarrierType


class EdgeElData(EdgeData):
    resistance: float = 0  # ohm per km
    reactance: float = 0  # ohm per km
    # Voltage for line (single value) or transformer (tuple)
    voltage: Union[float, Tuple[float, float]] = None  # kV.
    # Power loss in MW as function of power transfer in MW:
    power_loss_function: Optional[Tuple[List[float], List[float]]] = None
    carrier: CarrierType = CarrierType.EL


class EdgeHeatData(EdgeData):
    # Heat loss in MW as function of energy transfer in MW:
    power_loss_function: Optional[Tuple[List[float], List[float]]] = None
    carrier: CarrierType = CarrierType.HEAT


class EdgeHydrogenData(EdgeData):
    bidirectional: bool = False
    carrier: CarrierType = CarrierType.HYDROGEN


class EdgeFluidData(EdgeData):
    # wellstream, oil, water, gas
    pressure_from: float = None
    pressure_to: float = None
    diameter_mm: float = None
    temperature_K: float = None
    height_m: float = 0
    num_pipes: Optional[int] = None
    bidirectional: bool = False
    # allovable relative deviation of pressure from nominal values,
    pressure_from_maxdeviation: Optional[float] = None
    pressure_to_maxdeviation: Optional[float] = None
    carrier: CarrierType = CarrierType.FLUID


class EdgeGasData(EdgeFluidData):
    carrier: CarrierType = CarrierType.GAS


class EdgeOilData(EdgeFluidData):
    carrier: CarrierType = CarrierType.OIL


class EdgeWellstreamData(EdgeFluidData):
    carrier: CarrierType = CarrierType.WELLSTREAM


class EdgeWaterData(EdgeFluidData):
    carrier: CarrierType = CarrierType.WATER



import logging

import numpy as np
import pandas as pd
import pyomo.environ as pyo

try:
    import matplotlib.pyplot as plt
    import plotly
    import plotly.express as px
    import plotly.subplots
    import plotly.io as pio
    import seaborn as sns
except ImportError:
    raise ImportError("In order to run this plotting module you need to install matplotlib, plotly and seaborn.")

sns.set_style("whitegrid")  # Optional: sns.set_palette("dark")

plotter = "plotly"  # matplotlib

logger = logging.getLogger(__name__)


def plot_df(df, id_var, filename=None, title=None, ylabel="value"):
    """Plot dataframe using plotly (saved to file)"""

    df_tidy = df.reset_index()
    df_tidy.rename(columns={0: ylabel}, inplace=True)
    fig = px.line(df_tidy, x="time", y=ylabel, color=id_var, title=title)

    # fig = px.line()
    # for cols in df:
    #    fig.add_scatter(df[cols])
    # fig = px.line(df,x=df.index,y=df.values,title=title)
    # py.iplot('data':[{'x':df.index,'y':df.values,'type':'line'}],
    #        'layout': {'title':'title'},
    #        filename=filename)
    if filename is not None:
        plotly.offline.plot(fig, filename=filename)

    return fig


def plot_device_profile(
    sim_result,
    optimisation_model,
    devs,
    filename=None,
    reverse_legend=True,
    include_forecasts=False,
    include_on_off=False,
    include_prep=False,
    devs_shareload=None,
):
    """plot forecast and actual profile (available power), and device output

    Parameters
    ==========
    sim_result : SimulationResults object
    optimisation_model : OptimisationModel object
    devs : list
        which devices to include
    devs_shareload : list ([]=ignore, None=do it for gas turbines)
        list of devices for which displayed power should be shared evenly
        (typically gas turbines)
        The optimision returns somewhat random distribution of load per device,
        in reality they will share load more or less evenly due to their
        frequency droop settings. Rather than imposing this in the optimisation,
        this is included in the plots. Default: gas turbine
    filename : Filename
    reverse_legend : Use reverse legend
    include_forecasts : Include forecasts
    include_on_off : To include on and off
    include_prep : To include prep

    """
    res = sim_result
    optimiser = optimisation_model
    if type(devs) is not list:
        devs = [devs]
    if include_forecasts & (len(devs) > 1):
        raise ValueError("Can only plot a single device when showing forecasts")
    df = res.device_flow.unstack(["carrier", "terminal"])[("el", "out")].unstack("device")
    if devs_shareload is None:
        # gas turbines:
        devs_shareload = [d for d, d_obj in optimiser.all_devices.items() if d_obj.dev_data.model == "gasturbine"]
    if devs_shareload:  # list is non-empty
        devs_online = (df[devs_shareload] > 0).sum(axis=1)
        devs_sum = df[devs_shareload].sum(axis=1)
        devs_mean = devs_sum / devs_online
        for c in devs_shareload:
            mask = df[c] > 0
            df.loc[mask, c] = devs_mean[mask]
    df.columns.name = "devices"
    df = df[devs]
    nrows = 1
    if include_on_off:
        nrows = nrows + 1
    if include_prep:
        nrows = nrows + 1
    df2 = res.device_is_on.unstack("device")[devs]
    df_prep = res.device_is_prep.unstack("device")[devs]
    timerange = list(res.device_is_on.index.get_level_values("time"))
    if plotter == "plotly":
        fig = plotly.subplots.make_subplots(rows=nrows, cols=1, shared_xaxes=True)
        colour = pio.templates[pio.templates.default].layout.colorway
        k = -1
        row_on_off = 2
        row_prep = 2
        if include_on_off:
            row_prep = 3
        for col in df:
            dev = col
            dev_data = optimiser.all_devices[dev].dev_data
            k = (k + 1) % len(colour)  # repeat colour cycle if necessary
            fig.add_scatter(
                x=df.index,
                y=df[col],
                line_shape="hv",
                name=col,
                line=dict(color=colour[k]),
                stackgroup="P",
                legendgroup=col,
                row=1,
                col=1,
            )
            if include_on_off & (dev_data.start_stop is not None):
                fig.add_scatter(
                    x=df2.index,
                    y=df2[col],
                    line_shape="hv",
                    name=col,
                    line=dict(color=colour[k], dash="dash"),
                    stackgroup="ison",
                    legendgroup=col,
                    row=row_on_off,
                    col=1,
                    showlegend=False,
                )
            if include_prep & (dev_data.start_stop is not None):
                fig.add_scatter(
                    x=df_prep.index,
                    y=df_prep[col],
                    line_shape="hv",
                    name=col,
                    line=dict(color=colour[k], dash="dash"),
                    stackgroup="ison",
                    legendgroup=col,
                    row=row_prep,
                    col=1,
                    showlegend=False,
                )
            if include_forecasts & (dev_data.profile is not None):
                curve = dev_data.profile
                device_P_max = dev_data.flow_max
                if curve in res.profiles_nowcast:
                    fig.add_scatter(
                        x=timerange,
                        y=res.profiles_nowcast.loc[timerange, curve] * device_P_max,
                        line_shape="hv",
                        line=dict(color=colour[k + 1]),
                        name="--nowcast",
                        legendgroup=col,
                        row=1,
                        col=1,
                    )
                fig.add_scatter(
                    x=timerange,
                    y=res.profiles_forecast.loc[timerange, curve] * device_P_max,
                    line_shape="hv",
                    line=dict(color=colour[k + 2]),
                    name="--forecast",
                    legendgroup=col,
                    row=1,
                    col=1,
                )
        fig.update_xaxes(row=1, col=1, title_text="")
        fig.update_xaxes(row=nrows, col=1, title_text="Timestep")
        fig.update_yaxes(row=1, col=1, title_text="Power supply (MW)")
        if include_on_off:
            fig.update_yaxes(row=row_on_off, col=1, title_text="On/off status")
        if include_prep:
            fig.update_yaxes(row=row_prep, col=1, title_text="Startup", nticks=2)
        if reverse_legend:
            fig.update_layout(legend_traceorder="reversed")
        fig.update_layout(height=600)
        # fig.show()
    elif plotter == "matplotlib":
        fig, axs = plt.subplots(nrows=nrows, ncols=1, shared_xaxes=True, figsize=(12, 1 + 3 * nrows))
        ax = axs[0]
        labels = []
        offset_online = 0
        # df.plot(ax=ax)
        for dev in devs:
            dev_data = optimiser.all_devices[dev].dev_data
            devname = "{}:{}".format(dev, dev_data.name)
            device_P_max = dev_data.flow_max
            df[dev].plot(ax=ax)
            # get the color of the last plotted line (the one just plotted)
            col = ax.get_lines()[-1].get_color()
            labels = labels + [devname]
            if include_forecasts & (dev_data.profile is not None):
                curve = dev_data.profile
                (res.profiles_nowcast.loc[timerange, curve] * device_P_max).plot(ax=ax, linestyle="--")
                # ax.set_prop_cycle(None)
                (res.profiles_forecast.loc[timerange, curve] * device_P_max).plot(ax=ax, linestyle=":")
                labels = labels + ["--nowcast", "--forecast"]
            if include_on_off & (dev_data.start_stop is not None):
                offset_online += 0.1
                df2[dev].plot(ax=ax, linestyle="--", color=col)
                labels = labels + ["--online"]
        plt.xlim(df.index.min(), df.index.max())
        ax.legend(labels, loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    else:
        raise ValueError(f"Plotter: {plotter} has not been implemented for plot device profile.")
    return fig


def plot_device_power_energy(sim_result, optimisation_model: pyo.Model, dev, filename=None, energy_fill_opacity=None):
    """Plot power in/out of device and storage level (if any)"""
    res = sim_result
    optimiser = optimisation_model
    dev_data = optimiser.all_devices[dev].dev_data
    device_name = "{}:{}".format(dev, dev_data.name)

    if dev_data.model == "storagehydrogen":  # Fixme: replace with isinstance class type
        # isinstance(dev_data.model,oogeso.dto.DeviceStorageHydrogenData)
        carrier = "hydrogen"
        flow_title = "Flow (Sm3/s)"
        energy_storage_title = "Energy storage( Sm3)"
    else:
        carrier = "el"
        flow_title = "Power (MW)"
        energy_storage_title = "Energy storage (MWh)"
    # Power flow in/out
    df_flow = res.device_flow[dev, carrier].unstack("terminal")
    if res.device_storage_energy is None:
        df_storage_energy = pd.DataFrame()
    else:
        df_storage_energy = res.device_storage_energy.unstack("device")
    if dev in df_storage_energy:
        df_storage_energy = df_storage_energy[dev]
        df_storage_energy.index = df_storage_energy.index + 1

    if plotter == "plotly":
        fig = plotly.subplots.make_subplots(specs=[[{"secondary_y": True}]])
        for col in df_flow.columns:
            fig.add_scatter(
                x=df_flow.index,
                y=df_flow[col],
                line_shape="hv",
                name=col,
                secondary_y=True,
                fill="tozeroy",
            )
        if not df_storage_energy.empty:
            fig.add_scatter(
                x=df_storage_energy.index,
                y=df_storage_energy,
                name="storage",
                secondary_y=False,
                fill="tozeroy",
            )  # ,line=dict(dash='dot'))
            if energy_fill_opacity is not None:
                k = len(fig["data"]) - 1
                linecol = plotly.colors.DEFAULT_PLOTLY_COLORS[k]
                opacity = energy_fill_opacity
                fillcol = "rgba({}, {})".format(linecol[4:][:-1], opacity)
                fig["data"][k]["fillcolor"] = fillcol
                fig["data"][k]["fill"] = "tozeroy"
            fig.update_yaxes(title_text=energy_storage_title, secondary_y=False, side="right")
        fig.update_xaxes(title_text="Timestep")
        fig.update_yaxes(title_text=flow_title, secondary_y=True, side="left")

    elif plotter == "matplotlib":
        fig = plt.figure(figsize=(12, 4))
        plt.title(device_name)
        ax = plt.gca()
        df_flow.plot(ax=ax, drawstyle="steps-post", marker=".")
        ax.set_xlabel("Timestep")
        ax.set_ylabel(flow_title)
        tmin = df_flow.index.get_level_values("time").min()
        tmax = df_flow.index.get_level_values("time").max() + 1
        ax.set_ylim(0, dev_data.flow_max)
        ax.legend(loc="upper left")  # , bbox_to_anchor =(1.01,0),frameon=False)

        if not df_storage_energy.empty:
            ax2 = ax.twinx()
            ax2.grid(None)
            df_storage_energy.plot(ax=ax2, linestyle=":", color="black")
            ax2.set_ylabel("Energy (MWh)")  # ,color="red")
            if dev_data.model in ["storage_el"]:
                ax2.set_ylim(0, dev_data.max_E)
            elif dev_data.model in ["well_injection"]:
                ax2.set_ylim(-dev_data.max_E / 2, dev_data.max_E / 2)
            # ax2.tick_params(axis='y', labelcolor="red")
            ax2.legend(loc="upper right")
        ax.set_xlim(tmin, tmax)
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    else:
        raise ValueError(f"Plotter: {plotter} has not been implemented for plot device power energy.")
    return fig


def plot_sum_power_mix(
    sim_result,
    optimisation_model,
    carrier,
    filename=None,
    reverse_legend=True,
    exclude_zero=False,
    devs_shareload=None,
):
    """
    Plot power mix

    Parameters
    ==========
    sim_result : SimulationResult object
    optimisation_model : Oogeso optimisation model
    carrier : string
    devs_shareload : list ([]=ignore, None=do it for gas turbines)
        list of devices for which power should be shared evenly (typically gas turbines)
        The optimision returns somewhat random distribution of load per device,
        in reality they will share load more or less evenly due to their
        frequency droop settings. Rather than imposing this in the optimisation,
        this is included in the plots.
    filename : Name of file
    reverse_legend : Use reverse legend
    exclude_zero : To exclude zero values
    """
    # optimiser = simulator.optimiser
    res = sim_result
    optimiser = optimisation_model
    # Power flow in/out
    df_flow = res.device_flow
    t_min = df_flow.index.get_level_values("time").min()
    t_max = df_flow.index.get_level_values("time").max() + 1
    mask_carrier = df_flow.index.get_level_values("carrier") == carrier
    mask_in = df_flow.index.get_level_values("terminal") == "in"
    mask_out = df_flow.index.get_level_values("terminal") == "out"
    df_flow_out = df_flow[mask_carrier & mask_out]
    df_flow_out.index = df_flow_out.index.droplevel(level=("carrier", "terminal"))
    df_flow_out = df_flow_out.unstack(0)
    columns_to_keep = optimiser.get_devices_in_out(carrier_out=carrier)
    df_flow_out = df_flow_out[columns_to_keep]
    df_flow_in = df_flow[mask_carrier & mask_in]
    df_flow_in.index = df_flow_in.index.droplevel(level=("carrier", "terminal"))
    df_flow_in = df_flow_in.unstack(0)
    columns_to_keep = optimiser.get_devices_in_out(carrier_in=carrier)
    df_flow_in = df_flow_in[columns_to_keep]

    if (devs_shareload is None) and (carrier in ["el", "heat"]):
        # gas turbines:
        devs_shareload = [d for d, d_obj in optimiser.all_devices.items() if d_obj.dev_data.model == "gasturbine"]
        logger.debug("Shared load=%s", devs_shareload)
    if devs_shareload:  # list is non-empty
        devs_online = (df_flow_out[devs_shareload] > 0).sum(axis=1)
        devs_sum = df_flow_out[devs_shareload].sum(axis=1)
        devs_mean = devs_sum / devs_online
        for c in devs_shareload:
            mask = df_flow_out[c] > 0
            df_flow_out.loc[mask, c] = devs_mean[mask]

    if exclude_zero:
        df_flow_in = df_flow_in.loc[:, df_flow_in.sum() != 0]
        df_flow_out = df_flow_out.loc[:, df_flow_out.sum() != 0]

    if plotter == "plotly":
        fig = plotly.subplots.make_subplots(rows=2, cols=1, shared_xaxes=True, vertical_spacing=0.05)
        for col in df_flow_in:
            fig.add_scatter(
                x=df_flow_in.index,
                y=df_flow_in[col],
                line_shape="hv",
                name="in:" + col,
                stackgroup="in",
                legendgroup=col,
                row=2,
                col=1,
            )
        for col in df_flow_out:
            fig.add_scatter(
                x=df_flow_out.index,
                y=df_flow_out[col],
                line_shape="hv",
                name="out:" + col,
                stackgroup="out",
                legendgroup=col,
                row=1,
                col=1,
            )
        fig.update_xaxes(row=1, col=1, title_text="")
        fig.update_xaxes(row=2, col=1, title_text="Timestep")
        fig.update_yaxes(row=1, col=1, title_text="Power supply (MW)")
        fig.update_yaxes(row=2, col=1, title_text="Power consumption (MW)")
        if reverse_legend:
            fig.update_layout(legend_traceorder="reversed")
        fig.update_layout(height=600)
        # fig.show()
    elif plotter == "matplotlib":
        fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(12, 8))
        plt.suptitle("Sum power ({})".format(carrier))
        df_flow_out.plot.area(ax=axes[0], linewidth=0)
        df_flow_in.plot.area(ax=axes[1], linewidth=0)
        axes[0].set_ylabel("Power supply (MW)")
        axes[1].set_ylabel("Power consumption (MW)")
        axes[0].set_xlabel("")
        axes[1].set_xlabel("Timestep")
        for ax in axes:
            if reverse_legend:
                handles, labels = ax.get_legend_handles_labels()
                ax.legend(
                    handles[::-1],
                    labels[::-1],
                    loc="lower left",
                    bbox_to_anchor=(1.01, 0),
                    frameon=False,
                )
            else:
                ax.legend(loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
            ax.set_xlim(t_min, t_max)

        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    else:
        raise ValueError(f"Plotter: {plotter} has not been implemented for plot Power Mix.")
    return fig


def plot_export_revenue(sim_result, filename=None, currency="$"):
    export_revenue = sim_result.export_revenue.unstack("carrier")
    if plotter == "plotly":
        dfplot = export_revenue.loc[:, export_revenue.sum() > 0]
        fig = px.area(dfplot)
        fig.update_xaxes(title_text="Timestep")
        fig.update_yaxes(title_text="Revenue ({}/s)".format(currency))
    elif plotter == "matplotlib":
        fig = plt.figure(figsize=(12, 4))
        plt.title("Export revenue ($/s)")
        ax = plt.gca()
        ax.set_ylabel("{}/s".format(currency))
        ax.set_xlabel("Timestep")
        (export_revenue.loc[:, export_revenue.sum() > 0]).plot.area(ax=ax, linewidth=0)
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    else:
        raise ValueError(f"Plotter: {plotter} has not been implemented for plot export revenue.")
    return fig


def plot_CO2_rate(sim_result, filename=None):
    plt.figure(figsize=(12, 4))
    plt.title("CO2 emission rate (kgCO2/s)")
    ax = plt.gca()
    ax.set_ylabel("kgCO2/s")
    ax.set_xlabel("Timestep")
    sim_result.co2_rate.plot()
    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")


def plot_CO2_rate_per_device(
    sim_result,
    optimisation_model,
    filename=None,
    reverse_legend=False,
    device_shareload=None,
):

    dfco2rate = sim_result.co2_rate_per_dev.unstack("device")
    all_devices = optimisation_model.all_devices
    dfplot = dfco2rate.loc[:, ~(dfco2rate == 0).all()].copy()

    if device_shareload is None:
        # gas turbines:
        device_shareload = [d for d, d_obj in all_devices.items() if d_obj.dev_data.model == "gasturbine"]

    if device_shareload:  # list is non-empty
        device_shareload = [d for d in device_shareload if d in dfplot]
        devs_online = (dfplot[device_shareload] > 0).sum(axis=1)
        devs_sum = dfplot[device_shareload].sum(axis=1)
        devs_mean = devs_sum / devs_online
        for c in device_shareload:
            mask = dfplot[c] > 0
            dfplot.loc[mask, c] = devs_mean[mask]

    if plotter == "plotly":
        fig = plotly.subplots.make_subplots(rows=1, cols=1)
        for col in dfplot:
            fig.add_scatter(
                x=dfplot.index,
                y=dfplot[col],
                line_shape="hv",
                name=col,
                stackgroup="one",
                row=1,
                col=1,
            )
        fig.update_xaxes(title_text="Timestep")
        fig.update_yaxes(title_text="Emission rate (kgCO2/s)")
        if reverse_legend:
            fig.update_layout(legend_traceorder="reversed")
        fig.update_layout(height=600)
    else:
        fig = plt.figure(figsize=(12, 4))
        ax = plt.gca()
        ax.set_ylabel("Emission rate (kgCO2/s)")
        ax.set_xlabel("Timestep")
        dfplot.plot.area(ax=ax, linewidth=0)
        if reverse_legend:
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(
                handles[::-1],
                labels[::-1],
                loc="lower left",
                bbox_to_anchor=(1.01, 0),
                frameon=False,
            )
        else:
            ax.legend(loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)

        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    return fig


def plot_CO2_intensity(sim_result, filename=None):
    title = "CO2 intensity (kgCO2/Sm3oe)"
    x_label = "Timestep"
    y_label = "CO2 intensity (kgCO2/Sm3oe)"
    df_plot = sim_result.co2_intensity
    if plotter == "plotly":
        fig = px.line(df_plot, x=df_plot.index, y=df_plot.values)  # ,title=title)
        fig.update_xaxes(title_text=x_label)
        fig.update_yaxes(title_text=y_label)
    else:
        fig = plt.figure(figsize=(12, 4))
        plt.title(title)
        ax = plt.gca()
        ax.set_ylabel(y_label)
        ax.set_xlabel(x_label)
        df_plot.plot()
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    return fig


def plot_profiles(profiles, filename=None):
    """Plot profiles (forecast and actual)"""
    fig = None
    if isinstance(profiles, list):
        # list of TimeSeriesData objects
        df = pd.DataFrame()
        for d in profiles:
            df[(d.id, "forecast")] = d.data
            if d.data_nowcast is not None:
                df[(d.id, "nowcast")] = d.data_nowcast
        df.index.name = "timestep"
        df.columns = pd.MultiIndex.from_tuples(df.columns, names=("variable", "type"))
    elif isinstance(profiles, dict):
        df = pd.concat(
            {
                "nowcast": profiles["actual"],
                "forecast": profiles["forecast"],
            }
        )
        df.index.names = ["type", "timestep"]
    else:
        raise Exception("profiles must be input data format or internal format")
    if plotter == "plotly":
        df = df.reset_index()
        df = df.melt(id_vars=["timestep"])
        fig = px.line(
            df,
            x="timestep",
            y="value",
            line_group="type",
            color="variable",
            line_dash="type",
            line_shape="hv",
        )
    elif plotter == "matplotlib":
        plt.figure(figsize=(12, 4))
        ax = plt.gca()
        df.loc[:, df.columns.get_level_values("type") == "forecast"].plot(ax=ax)
        # reset color cycle (so using the same as for the actual plot):
        ax.set_prop_cycle(None)
        df.loc[:, df.columns.get_level_values("type") == "nowcast"].plot(ax=ax, linestyle=":")
        ax.legend(loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
        plt.xlabel("Timestep")
        plt.ylabel("Relative value")
        plt.title("Forecast and nowcast profile (forecast=sold line)")
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    return fig


def plot_device_power_flow_pressure(sim_result, optimisation_model: pyo.Model, dev, carriers_inout=None, filename=None):
    res = sim_result
    all_devices = optimisation_model.all_devices
    dev_obj = all_devices[dev]
    dev_data = dev_obj.dev_data
    node = dev_data.node_id
    devname = "{}:{}".format(dev, dev_data.name)
    if carriers_inout is None:
        carriers_inout = {"in": dev_obj.carrier_in, "out": dev_obj.carrier_out}
    if "serial" in carriers_inout:
        del carriers_inout["serial"]

    plt.figure(figsize=(12, 4))
    ax = plt.gca()
    for inout, carriers in carriers_inout.items():
        if inout == "in":
            ls = "--"
        else:
            ls = ":"
        for carr in carriers:
            ax.plot(
                res.device_flow.unstack([0, 1, 2])[(dev, carr, inout)],
                ls,
                label="DeviceFlow ({},{})".format(carr, inout),
            )
    # Pressure
    for inout, carriers in carriers_inout.items():
        for carr in carriers:
            if carr != "el":
                ax.plot(
                    res.terminal_pressure.unstack(0)[node].unstack([0, 1])[(carr, inout)],
                    label="TerminalPressure ({},{})".format(carr, inout),
                )
    plt.title(devname)
    plt.xlabel("Timestep")
    plt.ylabel("Value")
    plt.legend(loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")


def plot_gas_turbine_efficiency(
    fuel_A=2.35, fuel_B=0.53, energy_content=40, co2_content=2.34, filename=None, P_max=None
):
    """
    co2content : CO2 content, kgCO2/Sm3gas
    energycontent: energy content, MJ/Sm3gas
    A,B : linear parameters
    """

    x_pow = np.linspace(0, 1, 50)
    y_fuel = fuel_B + fuel_A * x_pow

    nplots = 3 if P_max is None else 4
    plt.figure(figsize=(4 * nplots, 4))

    if P_max is not None:
        y_fuel_sm3 = y_fuel * P_max / energy_content  # Sm3/s
        plt.subplot(1, nplots, 1)
        plt.title("Fuel usage (Sm3/h)")
        plt.xlabel("Electric power output (MW)")
        plt.plot(x_pow * P_max, y_fuel_sm3 * 3600)  # per hour
        plt.ylim(bottom=0)

    plt.subplot(1, nplots, nplots - 2)
    plt.title("Fuel usage ($P_{gas}/P_{el}^{max}$)")
    plt.xlabel("Electric power output ($P_{el}/P_{el}^{max}$)")
    # plt.ylabel("Gas power input ($P_{gas}/P_{el}^{max}$)")
    plt.plot(x_pow, y_fuel)
    plt.ylim(bottom=0)

    plt.subplot(1, nplots, nplots - 1)
    plt.title("Efficiency ($P_{el}/P_{gas}$)")
    plt.xlabel("Electric power output ($P_{el}/P_{el}^{max}$)")
    plt.plot(x_pow, x_pow / y_fuel)

    #    plt.subplot(1,3,3)
    #    plt.title("Specific fuel usage ($P_{gas}/P_{el}$)")
    #    plt.xlabel("Electric power output ($P_{el}/P_{el}^{max}$)")
    #    plt.plot(x_pow,y_fuel/x_pow)
    #    plt.ylim(top=30)

    # 1 MJ = 3600 MWh
    with np.errstate(divide="ignore", invalid="ignore"):
        emissions = 3600 * co2_content / energy_content * y_fuel / x_pow
    plt.subplot(1, nplots, nplots)
    plt.title("Emission intensity (kgCO2/MWh)")
    plt.xlabel("Electric power output ($P_{el}/P_{el}^{max}$)")
    plt.plot(x_pow, emissions)
    plt.ylim(top=2000)

    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")


def plot_reserve(
    sim_result,
    optimisation_model,
    include_margin=True,
    dynamic_margin=True,
    use_forecast=False,
    include_sum=True,
    devs_shareload=None,
):
    """Plot unused online capacity by all el devices
    devs_shareload : list ([]=ignore, None=do it for gas turbines)
        list of devices for which power should be shared evenly (typically gas turbines)
        The optimision returns somewhat random distribution of load per device,
        in reality they will share load more or less evenly due to their
        frequency droop settings. Rather than imposing this in the optimisation,
        this is included in the plots.
    """
    df_devs = pd.DataFrame()
    res = sim_result
    optimiser = optimisation_model
    timerange = list(res.el_reserve.index)
    margin_incr = pd.DataFrame(0, index=timerange, columns=["margin"])
    if devs_shareload is None:
        # split load evently amongst gas turbines:
        devs_shareload = [d for d, d_obj in optimiser.all_devices.items() if d_obj.dev_data.model == "gasturbine"]
        logger.debug("Shared load=%s", devs_shareload)
    if devs_shareload:  # list is non-empty
        devs_online = res.device_is_on[devs_shareload].unstack("device").sum(axis=1)
        devs_sum = res.device_flow.unstack("device").loc[("el", "out")][devs_shareload].sum(axis=1)
        devs_mean = devs_sum / devs_online
    for d, dev_obj in optimiser.all_devices.items():
        dev_data = dev_obj.dev_data
        device_model = dev_data.model
        rf = 1
        if "el" in dev_obj.carrier_out:
            # Generators and storage
            max_value = dev_data.flow_max
            if dev_data.profile is not None:
                ext_profile = dev_data.profile
                if use_forecast or (ext_profile not in res.profiles_nowcast):
                    max_value = max_value * res.profiles_forecast.loc[timerange, ext_profile]
                else:
                    max_value = max_value * res.profiles_nowcast.loc[timerange, ext_profile]
            if dev_data.start_stop is not None:  # device_model in ["gasturbine"]:
                is_on = res.device_is_on[d]
                max_value = is_on * max_value
            elif device_model in ["storage_el"]:
                max_value = res.dfDeviceStoragePmax[d] + res.device_flow[d, "el", "in"]
            if dev_data.reserve_factor is not None:
                reserve_factor = dev_data.reserve_factor
                if reserve_factor == 0:
                    # device does not count towards reserve
                    rf = 0
                if dynamic_margin:
                    # instead of reducing reserve, increase the margin instead
                    # R*0.8-M = R - (M+0.2R) - this shows better in the plot what
                    margin_incr["margin"] += rf * max_value * (1 - reserve_factor)
                else:
                    max_value = max_value * reserve_factor
            cap_avail = rf * max_value
            p_generating = rf * res.device_flow[d, "el", "out"]
            if d in devs_shareload:
                # plot evenly distributed output (devs_mean)
                p_generating = rf * res.device_is_on[d] * devs_mean
            reserv = cap_avail - p_generating
            df_devs[d] = reserv
    df_devs.columns.name = "device"
    if plotter == "plotly":
        fig = px.area(df_devs, line_shape="hv")
        if include_sum:
            fig.add_scatter(
                x=df_devs.index,
                y=df_devs.sum(axis=1),
                name="SUM",
                line=dict(dash="dot", color="black"),
                line_shape="hv",
            )
        if include_margin:
            margin = optimiser.all_networks["el"].carrier_data.el_reserve_margin
            # wind contribution (cf compute reserve)
            margin_incr["margin"] = margin_incr["margin"] + margin
            fig.add_scatter(
                x=margin_incr.index,
                y=margin_incr["margin"],
                name="Margin",
                line=dict(dash="dot", color="red"),
                mode="lines",
            )
        fig.update_xaxes(title_text="Timestep")
        fig.update_yaxes(title_text="Reserve (MW)")
        #    return df_devs,margin_incr
    elif plotter == "matplotlib":
        ax = df_devs.plot.area()
        if include_sum:
            df_devs.sum(axis=1).plot(style=":", color="black", drawstyle="steps-post")
        if include_margin:
            margin = optimiser.all_networks["el"].carrier_data.el_reserve_margin
            # wind contribution (cf compute reserve)
            margin_incr["margin"] = margin_incr["margin"] + margin
            margin_incr[["margin"]].plot(style=":", color="red", ax=ax)

        plt.xlabel("Timestep")
        plt.ylabel("Reserve (MW)")
        fig = plt.gcf()
    else:
        raise ValueError(f"Plotter: {plotter} has not been implemented for plot reserve.")
    return fig


def plot_el_backup(sim_result, filename=None, show_margin=False, return_margin=False):
    """plot reserve capacity vs device power output"""
    res = sim_result
    res_dev = res.el_backup.unstack("device")
    df_device_flow = res.device_flow.copy()
    carrier = "el"
    mask_carrier = df_device_flow.index.get_level_values("carrier") == carrier
    mask_out = df_device_flow.index.get_level_values("terminal") == "out"
    df_device_flow.index = df_device_flow.index.droplevel(level=("carrier", "terminal"))
    df_device_flow = df_device_flow[mask_carrier & mask_out].unstack(0)
    df_device_flow = df_device_flow[res_dev.columns]
    df_margin = (res_dev - df_device_flow).min(axis=1)
    if plotter == "plotly":
        fig = px.line()  # title="Online backup capacity (solid lines) vs device output (dotted lines)")
        colour = plotly.colors.DEFAULT_PLOTLY_COLORS
        k = 0
        for col in res_dev:
            fig.add_scatter(
                x=res_dev.index,
                y=res_dev[col],
                mode="lines",
                legendgroup=col,
                name="{} R_other".format(col),
                line_shape="hv",
                line=dict(color=colour[k]),
            )
            fig.add_scatter(
                x=df_device_flow.index,
                y=df_device_flow[col],
                legendgroup=col,
                name="{} P_out".format(col),
                line_shape="hv",
                line=dict(color=colour[k], dash="dot"),
            )
            k = k + 1
        if show_margin:
            fig.add_scatter(
                x=df_margin.index,
                y=df_margin,
                name="MARGIN",
                line=dict(color="black"),
                line_shape="hv",
            )
        fig.update_xaxes(title_text="Timestep")
        fig.update_yaxes(title_text="Power (MW)")
    elif plotter == "matplotlib":
        fig = plt.figure(figsize=(12, 4))
        ax = plt.gca()
        res_dev.plot(ax=ax, legend=True, alpha=1, linestyle="-")
        labels = list(res_dev.columns)
        if show_margin:
            df_margin.plot(ax=ax, linestyle="-", linewidth=3, color="black", label="MARGIN")
            labels = labels + ["MARGIN"]
        plt.gca().set_prop_cycle(None)
        df_device_flow.plot(ax=ax, linestyle=":", legend=False, alpha=1)
        plt.title("Online backup capacity (solid lines) vs device output (dotted lines)")
        ax.legend(labels, loc="lower left", bbox_to_anchor=(1.01, 0), frameon=False)
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
    else:
        raise ValueError(f"Plotter: {plotter} has not been implemented for plot el backup.")
    if return_margin:
        return df_margin

    return fig



import json
import logging
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from pydantic import BaseModel

from oogeso import dto
from oogeso.dto.mapper import (
    get_carrier_data_class_from_str,
    get_device_data_class_from_str,
    get_edge_data_class_from_str,
)

logger = logging.getLogger(__name__)


class DataclassJSONEncoder(json.JSONEncoder):
    """Serialize (save to file)"""

    def default(self, obj: Any):
        if isinstance(obj, BaseModel):
            return obj.dict()
        if isinstance(obj, pd.Series):
            dct = obj.reset_index().to_json(orient="records")
            return dct
        if isinstance(obj, pd.DataFrame):
            dct = obj.reset_index().to_json(orient="records")
            return dct
        if isinstance(obj, BaseModel):
            return obj.dict()
        return super().default(obj)


class DataclassJSONDecoder(json.JSONDecoder):
    """Deserialize (read from file)"""

    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    @staticmethod
    def _new_carrier(dct: Dict[str, str]):
        model = dct["id"]
        carrier_class = get_carrier_data_class_from_str(model_name=model)
        return carrier_class(**dct)

    @staticmethod
    def _new_node(dct: Dict[str, str]) -> dto.NodeData:
        return dto.NodeData(**dct)

    @staticmethod
    def _new_edge(dct: Dict[str, str]) -> dto.EdgeData:
        carrier = dct.pop("carrier")  # gets and deletes model from dictionary
        edge_class = get_edge_data_class_from_str(carrier_name=carrier)
        return edge_class(**dct)

    @staticmethod
    def _new_device(dct: Dict[str, Union[str, object]]) -> dto.DeviceData:
        # logger.debug(dct)
        model = dct.pop("model")  # gets and deletes model from dictionary
        logger.debug(model)
        start_stop: Optional[Dict] = dct.pop("start_stop", None)
        if start_stop is not None:
            startstop_obj = dto.StartStopData(**start_stop)
            dct["start_stop"] = startstop_obj
        dev_class = get_device_data_class_from_str(model_name=model)
        return dev_class(**dct)

    @staticmethod
    def _new_profile(dct: Dict[str, Union[str, List[float]]]) -> dto.TimeSeriesData:
        name = dct["id"]
        data = dct["data"]
        data_nowcast = None
        if "data_nowcast" in dct:
            data_nowcast = dct["data_nowcast"]
        return dto.TimeSeriesData(id=name, data=data, data_nowcast=data_nowcast)

    def object_hook(self, dct):  # noqa
        carriers = []
        nodes = []
        edges = []
        # devs = {}
        devs = []
        profiles = []
        # profiles_nowcast = {}
        if "nodes" in dct:
            # Top level
            d_params = dct["parameters"]
            params = dto.OptimisationParametersData(**d_params)
            for n in dct["carriers"]:
                carriers.append(self._new_carrier(n))
            for n in dct["nodes"]:
                nodes.append(self._new_node(n))
            for n in dct["edges"]:
                edges.append(self._new_edge(n))
            for n in dct["devices"]:
                devs.append(self._new_device(n))
            if "profiles" in dct:
                for n in dct["profiles"]:
                    profiles.append(self._new_profile(n))
                    # profiles[i] = self._new_profile(n)
            # if "profiles_nowcast" in dct:
            #    for i, n in dct["profiles_nowcast"].items():
            #        profiles_nowcast[i] = self._new_profile(n)
            energy_system_data = dto.EnergySystemData(
                carriers=carriers,
                nodes=nodes,
                edges=edges,
                devices=devs,
                parameters=params,
                profiles=profiles,
                # profiles_nowcast=profiles_nowcast,
            )
            return energy_system_data
        return dct


def serialize_oogeso_data(energy_system_data: dto.EnergySystemData):
    json_string = json.dumps(energy_system_data, cls=DataclassJSONEncoder, indent=2)
    return json_string


def deserialize_oogeso_data(json_data):
    energy_system_data = json.loads(json_data, cls=DataclassJSONDecoder)
    return energy_system_data


# Deserialize result object
class OogesoResultJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, dct):  # noqa
        res_dfs = {}
        # profiles_nowcast = {}
        if "device_flow" in dct:
            # Top level
            for k, val in dct.items():
                logger.debug(k)
                if val is None:
                    res_dfs[k] = None
                else:
                    val2 = json.loads(val)
                    df = pd.DataFrame.from_dict(val2)
                    if df.empty:
                        df = None
                    elif k in ["profiles_forecast", "profiles_nowcast"]:
                        df = df.set_index("index")
                    else:
                        # change to multi-index pandas Series
                        index_names = ["device", "time", "carrier", "terminal", "edge", "node"]
                        multiind = [c for c in df.columns if c in index_names]
                        df = df.set_index(multiind)
                        df = df.iloc[:, 0]
                    res_dfs[k] = df
            result_data = dto.SimulationResult(**res_dfs)
            return result_data
        return dct
